# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys
import logging

import cherrypy

import mini_buildd.misc
import mini_buildd.setup

LOG = logging.getLogger(__name__)


def log_init():
    """
    Setup CherryPy to use mini-buildd's logging mechanisms.
    """

    # listener
    def cherry_log(msg, level):
        LOG.log(level, msg)

    # subscribe to channel
    cherrypy.engine.subscribe('log', cherry_log)

    # turn off stderr/stdout logging
    cherrypy.log.screen = False

    # HTTP errors (status codes: 4xx-5xx)
    cherrypy.HTTPError.set_response = lambda msg: LOG.log(logging.ERROR, msg)


def shutdown():
    """
    Stop the CherryPy engine.
    """
    cherrypy.engine.exit()


def run(bind, wsgi_app):
    """
    Run the CherryPy WSGI Web Server.

    :param bind: the bind address to use.
    :type bind: string
    :param wsgi_app: the web application to process.
    :type wsgi_app: WSGI-application

    """
    def add_static_handler(directory, root, path):
        "Shortcut to add a static handler."
        cherrypy.tree.mount(
            cherrypy.tools.staticdir.handler(
                section="/",
                dir=directory,
                root=root,
                content_types={"log": "text/plain; charset=UTF-8", "buildlog": "text/plain; charset=UTF-8"}),
            path)

    log_init()

    cherrypy.config.update({'server.socket_host': mini_buildd.misc.HoPo(bind).host,
                            'server.socket_port': mini_buildd.misc.HoPo(bind).port,
                            'engine.autoreload_on': False})

    # Django: Add our own static directory
    add_static_handler(directory=".",
                       root="/usr/share/pyshared/mini_buildd/static",
                       path="/static")

    # Django: Add static support for the admin app
    add_static_handler(directory="static/admin",
                       root="/usr/lib/python{major}.{minor}/dist-packages/django/contrib/admin".format(major=sys.version_info[0], minor=sys.version_info[1]),
                       path="/static/admin")

    # Serve our Debian-installed html documentation directly
    add_static_handler(directory=".", root="/usr/share/doc/mini-buildd/html", path="/doc")

    # Serve repositories and log directories
    add_static_handler(directory=".", root=mini_buildd.setup.REPOSITORIES_DIR, path="/repositories")
    add_static_handler(directory=".", root=mini_buildd.setup.LOG_DIR, path="/log")

    # register wsgi app (django)
    cherrypy.tree.graft(wsgi_app)

    cherrypy.engine.start()
    cherrypy.engine.block()
